searchData={"items":[{"type":"module","title":"ProcessTree","doc":"`ProcessTree` is a module for avoiding global state in Elixir applications.","ref":"ProcessTree.html"},{"type":"module","title":"Motivation - ProcessTree","doc":"It's common in Elixir apps to rely on global references or global variables for locating application services or configuration parameters. This presents problems when we test our code. We need to work around our global data, for example by setting/resetting environment variables and using `async: false`.\n\n`ProcessTree` solves the problem by localizing data to a particular branch of the Elixir process tree. When testing with ExUnit, each ExUnit test process, and all processes spawned by the test process, can use `get/1` or `get/2` to see their own private copy of the data of interest.","ref":"ProcessTree.html#module-motivation"},{"type":"module","title":"How to use `get()` - ProcessTree","doc":"`get/1` and `get/2` can be used to make child processes \"see\" values stored in the process dictionaries of their ancestor processes. This is useful, for example, in ExUnit tests that spawn processes, such as tests that start LiveViews and GenServers.\n\nTo make data visible to child processes via `get/1` and `get/2`, we first put the data into the process dictionary of an ancestor process - in this case, an ExUnit test pid:\n\n``` elixir\ntest \"some test that starts a GenServer\" do\n  # ...\n  # add the data of interest to the process dictionary of the test pid\n  Process.put(:some_key, some_value)\n\n  # The GenServer process started here can use ProcessTree.get() to see\n  # the value we've bound to :some_key\n  server = MyGenserver.start_link()\n  # ...\nend\n```","ref":"ProcessTree.html#module-how-to-use-get"},{"type":"module","title":"Example use case - ProcessTree","doc":"[Customizing environment variables](./examples/environment-variable-example.md) in ExUnit tests while preserving `async: true`.","ref":"ProcessTree.html#module-example-use-case"},{"type":"module","title":"Smoothing over process ancestry complications - ProcessTree","doc":"OTP 25 [introduced](https://github.com/erlang/otp/pull/5768) the ability to find the parent of a process via `Process.info/2`. Prior to OTP 25, it was possible to find the parent of a process only for specific processes such as Task, GenServer, Agent, and Superivsor.\n\nEven under OTP 25+, `Process.info/2` is only useful for processes that are still alive, meaning it can't be used, for example, to find the grandparent of a process if the parent of the process has died. \n\n`ProcessTree` accounts for all complicating factors. Each of the functions exposed by `ProcessTree` will return the most complete answer possible, regardless of how the processes in the ancestry hierarchy are started or managed, regardless of which OTP version is in use, and so on.\n\nPut another way, `ProcessTree` is a \"no-judgments zone\". If you're following recommended guidelines for starting/running Elixir processes, `ProcessTree` will almost certainly meet your needs. But even in situations typically considered inadvisable or totally crazy, `ProcessTree` will do its very best to provide meaningful answers.","ref":"ProcessTree.html#module-smoothing-over-process-ancestry-complications"},{"type":"function","title":"ProcessTree.get/1","doc":"Starting with the calling process, recursively looks for a value for `key` in\nthe process dictionaries of the calling process and its known ancestors.\n\nIf a particular ancestor process has died, it looks up to the next ancestor, if possible.\n\nIf a non-nil value is found in the tree, the value is cached in the dictionary of\nthe calling process.\n\nReturns the first non-nil value found in the tree. If no value is found, returns `nil`.","ref":"ProcessTree.html#get/1"},{"type":"function","title":"ProcessTree.get/2","doc":"Starting with the calling process, recursively looks for a value for `key` in the\nprocess dictionaries of the calling process and its known ancestors.\n\nIf a particular ancestor process has died, it looks up to the next ancestor, if possible.\n\nIf a non-nil value is found in the tree, the value is cached in the dictionary of\nthe calling process.\n\nReturns the first non-nil value found in the tree.\n\nIf no value is found, the provided default value is cached in the dictionary of the\ncalling process and then returned.\n\nDefault values might typically be read from the Application environment. For example:\n\n```\ndefault_value = Application.get_env(:my_app, :some_key)\nvalue = ProcessTree.get(:some_key, default: default_value)\n```","ref":"ProcessTree.html#get/2"},{"type":"function","title":"ProcessTree.known_ancestors/1","doc":"Returns a list of the known ancestors of `pid`, from \"newest\" to \"oldest\".\n\nThe set of ancestors that is \"known\" depends on factors including:\n\n* The OTP major version the code is running under. (OTP 25 [introduced](https://github.com/erlang/otp/pull/5768)\n  new functionality for tracking ancestors.)\n* Whether the process and its ancestors are running in a supervision tree\n* Whether ancestor processes are still alive\n* Whether the given process and its ancestors were started via raw `spawn` or\n  were instead started as Tasks, Agents, GenServers or Supervisors\n\n`ProcessTree` takes these factors and more into account and produces the most complete\nlist of ancestors that is possible.\n\nIn the mainline case - running under a supervision tree, as recommended - `known_ancestors/1` will\nreturn a list that contains, at minimum, all of the ancestor Supervisors in the tree\nas well as the ancestor of the initial/topmost Supervisor.\n\nWhen running under OTP 25 and later, the list will also include all additional ancestors,\nup to and including the `:init` process (`PID `), provided that the additional ancestor\nprocesses are still alive.\n\nList items will be pids in all but the most unusual of circumstances. For example, if a GenServer\nis spawned by parent/grandparent GenServers that have registered names, and the parent GenServer dies,\nthen the parent & grandparent may be represented in the list of the child's known ancestors using\nthe their registered names - atoms, rather than pids. Precise behavior depends on OTP major version.","ref":"ProcessTree.html#known_ancestors/1"},{"type":"function","title":"ProcessTree.parent/1","doc":"Returns the parent of `pid`, if the parent is known.\n\nReturns `:unknown` if the parent is unknown.\n\nReturns `:undefined` if `pid` represents the `:init` process (`PID `).\n\nIf `pid` is part of a supervision tree, the parent will be known regardless of any other factors.\n\nIf the parent is known, the return value will be a pid in all but the most unusual of\ncircumstances. See `known_ancestors/1` for dicussion.","ref":"ProcessTree.html#parent/1"},{"type":"extras","title":"Environment Variable Example","doc":"# Environment Variable Example\n\n \n \n\nThis example demonstrates how `ProcessTree` can be used in tests to provide custom values for \nenvironment variables, while preserving `async: true`.\n\nThe example is a Phoenix app with a single LiveView that serves the site's index page. The \nLiveView uses an environment variable called `cutoff_date` to customize the appearance of the \nindex page. When a viewer hits the page before the cutoff date, they see this message:\n\n \n \n    Welcome! You've made the cutoff date :-)\n \n \n\nWhen a viewer hits the page after the cutoff date, they see this message:\n\n \n \n    Sorry! You've missed the cutoff date :-(\n \n \n\nFor \"real-life\" usage when `MIX_ENV=dev` or `MIX_ENV=prod`, the `cutoff_date` is set in `runtime.exs` as a hardcoded \nApplication environment variable:\n\n``` elixir\nconfig :environment_variable_example, cutoff_date: ~D[2024-01-01]\n```\n\nIn our tests of the LiveView, we override the value on a test-specific basis by writing the desired\ncustom value to the process dictionary of the ExUnit test pid:\n\n``` elixir\ntest \"when the cutoff date has not passed, shows a 'welcome' message\", %{conn: conn} do\n  tomorrow = Date.add(Date.utc_today(), 1)\n\n  # set the cutoff date to tomorrow by writing the value\n  # to the process dictionary of the ExUnit test pid\n  Process.put(:cutoff_date, tomorrow)\n\n  {:ok, _show_live, html} = live(conn, ~p\"/welcome\")\n\n  assert html =~ \"Welcome!\"\nend\n\ntest \"when cutoff date has passed, shows a 'sorry' message\", %{conn: conn} do\n  yesterday = Date.add(Date.utc_today(), -1)\n\n  # set the cutoff date to yesterday by writing the value\n  # to the process dictionary of the ExUnit test pid\n  Process.put(:cutoff_date, yesterday)\n\n  {:ok, _show_live, html} = live(conn, ~p\"/welcome\")\n\n  assert html =~ \"Sorry!\"\nend\n```\n\nSince this custom value is scoped to individual ExUnit tests, rather than to the global Application environment, the tests are safe for `async: true`.\n\nThe code for the LiveView is below. In the `cutoff_date()` function, the LiveView uses `ProcessTree` to look for a customized value. If no custom value is found (as when running in `dev` or `prod`), then the LiveView uses the `cutoff_date` retrieved from the Application environment, as configured in `runtime.exs` \n\n``` elixir\ndefmodule EnvironmentVariableExampleWeb.WelcomeLive.Show do\n  @moduledoc false\n  use EnvironmentVariableExampleWeb, :live_view\n\n  @impl true\n  def render(assigns) do\n    message = case Date.before?(Date.utc_today(), cutoff_date()) do\n      true ->\n        \"Welcome! You've made the cutoff date :-)\"\n\n      false ->\n        \"Sorry! You've missed the cutoff date :-(\"\n    end\n\n    assigns = assign(assigns, :message, message)\n\n    ~H\"\"\"\n       \n    \"\"\"\n  end\n\n  defp cutoff_date() do\n    # read the default cutoff_date from the Application environment\n    app_env_cutoff_date = Application.get_env(:environment_variable_example, :cutoff_date)\n\n    # ProcessTree will look for a customized cutoff_date value in the process dictionaries\n    # of this process and its ancestors.\n    #\n    # When running our tests, ProcessTree will find and return the customized value that\n    # we have inserted into the process dictionary of the ExUnit test pid, which is an\n    # ancestor of this LiveView process.\n    #\n    # When running in production, ProcessTree will not find a value in any ancestor\n    # dictionaries, and it will return the default value after caching it in the process\n    # dictionary of the current process.\n    ProcessTree.get(:cutoff_date, default: app_env_cutoff_date)\n  end\nend\n```\n\nDuring tests, the LiveView \"sees\" the custom value because `ProcessTree` looks up the process ancestry hierarchy to \nfind the value. The LiveView process is spawned by the ExUnit test pid (indirectly, via the ExUnit test supervisor), meaning that the test pid is an ancestor of the LiveView. `ProcessTree` eventually finds the custom value in the process dictionary of the test pid.\n\nIn production, `ProcessTree` returns the default value obtained from the Application environment.","ref":"environment-variable-example.html"}],"content_type":"text/markdown"}